/*-------------------------------------------------------------------------
 Local bus controller on SASEBO 
 
 File name   : ctrl_lbus.v
 Version     : Version 1.0
 Created     : MAR/05/2008
 Last update : AUG/08/2008
 Desgined by : Toshihiro Katashita
 
 
 Copyright (C) 2008 AIST
 
 By using this code, you agree to the following terms and conditions.
 
 This code is copyrighted by AIST ("us").
 
 Permission is hereby granted to copy, reproduce, redistribute or
 otherwise use this code as long as: there is no monetary profit gained
 specifically from the use or reproduction of this code, it is not sold,
 rented, traded or otherwise marketed, and this copyright notice is
 included prominently in any copy made.
 
 We shall not be liable for any damages, including without limitation
 direct, indirect, incidental, special or consequential damages arising
 from the use of this code.
 
 When you publish any results arising from the use of this code, we will
 appreciate it if you can cite our webpage
 (http://www.rcis.aist.go.jp/special/SASEBO/).
 -------------------------------------------------------------------------*/ 


//================================================ CTRL_LBUS
module CTRL_LBUS
  (lbus_a, lbus_di, lbus_do, lbus_wr, lbus_rd, // Local bus
   wd, we, ful, aful, rd, re, emp, aemp, rcnt, // Control signals of UART FIFO
   clk, rst);                                  // Clock and reset
   
   //------------------------------------------------
   // Local bus
   output [15:0] lbus_a;  // Address
   output [15:0] lbus_di; // Input data  (Controller -> Cryptographic module)
   output        lbus_wr; // Assert input data
   input [15:0]  lbus_do; // Output data (Cryptographic module -> Controller)
   output        lbus_rd; // Assert output data

   // Control signals of UART FIFO
   output [7:0]  wd;
   output        we;
   input         ful, aful;
   input [7:0]   rd;
   output        re;
   input         emp, aemp;
   input [11:0]  rcnt;

   // Clock and reset
   input         clk, rst;

   //------------------------------------------------
   reg [4:0]     state;
   reg [3:0]     cnt;
   reg [7:0]     wd;
   reg           we, re;
   wire          wr_busy, rd_busy;
   reg [15:0]    lbus_a;
   reg [15:0]    lbus_di;
   reg           lbus_rd, lbus_wr;
   reg [15:0]    fetched;
   
   //------------------------------------------------
   parameter [4:0]
                WAIT = 5'h00,
                CMD0 = 5'h01,
                CMD1 = 5'h02,
                RD0  = 5'h03,
                RD1  = 5'h04,
                RD2  = 5'h05,
                RD3  = 5'h06,
                RD4  = 5'h07,
                RD5  = 5'h08,
                RD6  = 5'h09,
                RD7  = 5'h0A,
                RD8  = 5'h0B,
                RD9  = 5'h0C,
                WR0  = 5'h0D,
                WR1  = 5'h0E,
                WR2  = 5'h0F,
                WR3  = 5'h10,
                WR4  = 5'h11,
                WR5  = 5'h12,
                WR6  = 5'h13,
                WR7  = 5'h14,
                WR8  = 5'h15,
                WR9  = 5'h16,
                TRAP = 5'h17;
                
   //------------------------------------------------
   assign wr_busy = ful | (aful & we);
   assign rd_busy = emp | (aemp & re);

   always @(posedge clk or posedge rst) begin
      if (rst) begin
         state <= WAIT; cnt <= 4'h0;
         we <= 0; wd <= 8'h0; re <= 0;
         lbus_a <= 16'h0; lbus_di <= 16'h0; lbus_rd <= 1; lbus_wr <= 1;
      end else
        case (state)
          WAIT: begin
             if (~rd_busy) begin state <= CMD0; re <= 1; end
             else          begin state <= WAIT; re <= 0; end
             we <= 0;
          end
          
          CMD0: begin state <= CMD1; re <= 0; end
          CMD1: begin
             if (rd==8'h00)      state <= RD0;
             else if (rd==8'h01) state <= WR0;
             else                state <= TRAP;
             re <= 0;
          end

          RD0: begin
             if (rcnt>=12'h2) begin state <= RD1; re <= 1; end
             else             begin state <= RD0; re <= 0; end
          end
          RD1: begin state <= RD2; re <= 1; end
          RD2: begin state <= RD3; re <= 0; lbus_a[15:8] <= rd; end
          RD3: begin state <= RD4; cnt <= 4'h0; lbus_a[7:0] <= rd; end
          RD4: begin
             if (cnt==4'h7) state <= RD5;
             else           state <= RD4;
             cnt <= cnt + 4'h1;
          end
          
          RD5: begin state <= RD6; cnt <= 4'h0; lbus_rd <= 0; end
          RD6: begin
             if (cnt==4'h7) state <= RD7;
             else           state <= RD6;
             cnt <= cnt + 4'h1;
          end

          RD7: begin state <= RD8; fetched <= lbus_do; lbus_rd <= 1; end
          RD8: begin
             if (~wr_busy) begin state <= RD9; we <= 1; end
             else          begin state <= RD8; we <= 0; end
             wd <= fetched[15:8];
          end
          RD9: begin
             if (~wr_busy) begin state <= WAIT; we <= 1; end
             else          begin state <= RD9;  we <= 0; end
             wd <= fetched[7:0];
          end
          
          WR0: begin
             if (rcnt>=12'h4) begin state <= WR1; re <= 1; end
             else             begin state <= WR0; re <= 0; end
          end
          WR1: begin state <= WR2; re <= 1; end
          WR2: begin state <= WR3; re <= 1; lbus_a[15:8] <= rd; end
          WR3: begin state <= WR4; re <= 1; lbus_a[7:0]  <= rd; end
          WR4: begin state <= WR5; re <= 0; lbus_di[15:8]<= rd; end
          WR5: begin state <= WR6; cnt <= 4'h0; lbus_di[7:0] <= rd; end
          WR6: begin
             if (cnt==4'h7) state <= WR7;
             else           state <= WR6;
             cnt <= cnt + 4'h1;
          end

          WR7: begin state <= WR8; cnt <= 4'h0; lbus_wr <= 0; end
          WR8: begin
             if (cnt==4'h7) state <= WR9;
             else           state <= WR8;
             cnt <= cnt + 4'h1;
          end
          
          WR9: begin state <= WAIT; lbus_wr <= 1; end

          TRAP: begin state <= TRAP; end
        endcase // case (state)
   end // always @ (posedge clk or posedge rst)
   
endmodule // CTRL_LBUS



//================================================ SW_INTERFACE
module SW_INTERFACE
  (wd, we, ful, aful, rd, re, emp, aemp, rcnt, 
   wd0, we0, ful0, aful0, rd0, re0, emp0, aemp0, rcnt0, 
   wd1, we1, ful1, aful1, rd1, re1, emp1, aemp1, rcnt1, 
   sel);

   //------------------------------------------------
   //
   input [7:0]   wd;
   input         we;
   output        ful, aful;
   output [7:0]  rd;
   input         re;
   output        emp, aemp;
   output [11:0] rcnt;
   // 
   output [7:0]  wd0,          wd1;
   output        we0,          we1;
   input         ful0, aful0,  ful1, aful1;
   input [7:0]   rd0,          rd1;
   output        re0,          re1;
   input         emp0, aemp0,  emp1, aemp1;
   input [11:0]  rcnt0,        rcnt1;
   //
   input         sel;
   
   //------------------------------------------------
   assign wd0  = (sel)? 8'h0  : wd;
   assign we0  = (sel)? 0     : we;
   assign wd1  = (sel)? wd    : 8'h0;
   assign we1  = (sel)? we    : 0;
   assign ful  = (sel)? ful1  : ful0;
   assign aful = (sel)? aful1 : aful0;

   assign rd   = (sel)? rd1   : rd0;
   assign re0  = (sel)? 0     : re;
   assign re1  = (sel)? re    : 0;
   assign emp  = (sel)? emp1  : emp0;
   assign aemp = (sel)? aemp1 : aemp0;
   assign rcnt = (sel)? rcnt1 : rcnt0;
   
endmodule // SW_INTERFACE
